@@ -16,6 +16,9 @@ import re
 import glob
 import os.path
 import subprocess
+import functools
+import itertools
+from pathlib import Path
 
 from .. import mlog
 from .. import coredata
@@ -46,11 +49,15 @@ from .compilers import (
     RunResult,
 )
 
+gnu_compiler_internal_libs = ('m', 'c', 'pthread', 'dl', 'rt')
+
 
 class CCompiler(Compiler):
+    # TODO: Replace this manual cache with functools.lru_cache
     library_dirs_cache = {}
     program_dirs_cache = {}
     find_library_cache = {}
+    internal_libs = gnu_compiler_internal_libs
 
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
         # If a child ObjC or CPP class has already set it, don't set it ourselves
@@ -140,6 +147,30 @@ class CCompiler(Compiler):
         '''
         return self.get_no_optimization_args()
 
+    def get_allow_undefined_link_args(self):
+        '''
+        Get args for allowing undefined symbols when linking to a shared library
+        '''
+        if self.id == 'clang':
+            if self.clang_type == compilers.CLANG_OSX:
+                # Apple ld
+                return ['-Wl,-undefined,dynamic_lookup']
+            else:
+                # GNU ld and LLVM lld
+                return ['-Wl,--allow-shlib-undefined']
+        elif self.id == 'gcc':
+            if self.gcc_type == compilers.GCC_OSX:
+                # Apple ld
+                return ['-Wl,-undefined,dynamic_lookup']
+            else:
+                # GNU ld and LLVM lld
+                return ['-Wl,--allow-shlib-undefined']
+        elif self.id == 'msvc':
+            # link.exe
+            return ['/FORCE:UNRESOLVED']
+        # FIXME: implement other linkers
+        return []
+
     def get_output_args(self, target):
         return ['-o', target]
 
@@ -168,42 +199,47 @@ class CCompiler(Compiler):
     def get_std_shared_lib_link_args(self):
         return ['-shared']
 
-    def get_library_dirs_real(self):
-        env = os.environ.copy()
-        env['LC_ALL'] = 'C'
-        stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=env)[1]
+    @functools.lru_cache()
+    def _get_search_dirs(self, env):
+        extra_args = ['--print-search-dirs']
+        stdo = None
+        with self._build_wrapper('', env, extra_args, None, 'compile', True) as p:
+            stdo = p.stdo
+        return stdo
+
+    @staticmethod
+    def _split_fetch_real_dirs(pathstr, sep=':'):
         paths = []
-        for line in stdo.split('\n'):
-            if line.startswith('libraries:'):
-                libstr = line.split('=', 1)[1]
-                paths = [os.path.realpath(p) for p in libstr.split(':')]
+        for p in pathstr.split(sep):
+            p = Path(p)
+            if p.exists():
+                paths.append(p.resolve().as_posix())
         return paths
 
-    def get_library_dirs(self):
-        key = tuple(self.exelist)
+    def get_compiler_dirs(self, env, name):
+        '''
+        Get dirs from the compiler, either `libraries:` or `programs:`
+        '''
+        stdo = self._get_search_dirs(env)
+        for line in stdo.split('\n'):
+            if line.startswith(name + ':'):
+                return CCompiler._split_fetch_real_dirs(line.split('=', 1)[1])
+        return []
+
+    def get_library_dirs(self, env):
+        key = (tuple(self.exelist), env)
         if key not in self.library_dirs_cache:
-            self.library_dirs_cache[key] = self.get_library_dirs_real()
+            self.library_dirs_cache[key] = self.get_compiler_dirs(env, 'libraries')
         return self.library_dirs_cache[key][:]
 
-    def get_program_dirs_real(self):
-        env = os.environ.copy()
-        env['LC_ALL'] = 'C'
-        stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=env)[1]
-        paths = []
-        for line in stdo.split('\n'):
-            if line.startswith('programs:'):
-                libstr = line.split('=', 1)[1]
-                paths = [os.path.realpath(p) for p in libstr.split(':')]
-        return paths
-
-    def get_program_dirs(self):
+    def get_program_dirs(self, env):
         '''
         Programs used by the compiler. Also where toolchain DLLs such as
         libstdc++-6.dll are found with MinGW.
         '''
-        key = tuple(self.exelist)
+        key = (tuple(self.exelist), env)
         if key not in self.program_dirs_cache:
-            self.program_dirs_cache[key] = self.get_program_dirs_real()
+            self.program_dirs_cache[key] = self.get_compiler_dirs(env, 'programs')
         return self.program_dirs_cache[key][:]
 
     def get_pic_args(self):
@@ -883,13 +919,13 @@ class CCompiler(Compiler):
 
     @classmethod
     def _get_trials_from_pattern(cls, pattern, directory, libname):
-        f = os.path.join(directory, pattern.format(libname))
+        f = Path(directory) / pattern.format(libname)
         # Globbing for OpenBSD
         if '*' in pattern:
             # NOTE: globbing matches directories and broken symlinks
             # so we have to do an isfile test on it later
-            return cls._sort_shlibs_openbsd(glob.glob(f))
-        return [f]
+            return cls._sort_shlibs_openbsd(glob.glob(str(f)))
+        return [f.as_posix()]
 
     @staticmethod
     def _get_file_from_list(files):
@@ -902,42 +938,30 @@ class CCompiler(Compiler):
         # First try if we can just add the library as -l.
         # Gcc + co seem to prefer builtin lib dirs to -L dirs.
         # Only try to find std libs if no extra dirs specified.
-        if not extra_dirs:
+        if not extra_dirs or libname in self.internal_libs:
             args = ['-l' + libname]
-            if self.links(code, env, extra_args=args):
+            largs = self.linker_to_compiler_args(self.get_allow_undefined_link_args())
+            if self.links(code, env, extra_args=(args + largs)):
                 return args
+            # Don't do a manual search for internal libs
+            if libname in self.internal_libs:
+                return None
         # Not found or we want to use a specific libtype? Try to find the
         # library file itself.
         patterns = self.get_library_naming(env, libtype)
-        for d in extra_dirs:
+        # Search in the specified dirs, and then in the system libraries
+        for d in itertools.chain(extra_dirs, self.get_library_dirs(env)):
             for p in patterns:
                 trial = self._get_trials_from_pattern(p, d, libname)
                 if not trial:
                     continue
+                # We just check whether the library exists. We can't do a link
+                # check because the library might have unresolved symbols that
+                # require other libraries.
                 trial = self._get_file_from_list(trial)
                 if not trial:
                     continue
                 return [trial]
-        # Search in the system libraries too
-        for d in self.get_library_dirs():
-            for p in patterns:
-                trial = self._get_trials_from_pattern(p, d, libname)
-                if not trial:
-                    continue
-                trial = self._get_file_from_list(trial)
-                if not trial:
-                    continue
-                # When searching the system paths used by the compiler, we
-                # need to check linking with link-whole, as static libs
-                # (.a) need to be checked to ensure they are the right
-                # architecture, e.g. 32bit or 64-bit.
-                # Just a normal test link won't work as the .a file doesn't
-                # seem to be checked by linker if there are no unresolved
-                # symbols from the main C file.
-                extra_link_args = self.get_link_whole_for([trial])
-                extra_link_args = self.linker_to_compiler_args(extra_link_args)
-                if self.links(code, env, extra_args=extra_link_args):
-                    return [trial]
         return None
 
     def find_library_impl(self, libname, env, extra_dirs, code, libtype):
@@ -1185,7 +1209,8 @@ class IntelCCompiler(IntelCompiler, CCompiler):
 class VisualStudioCCompiler(CCompiler):
     std_warn_args = ['/W3']
     std_opt_args = ['/O2']
-    ignore_libs = ('m', 'c', 'pthread')
+    ignore_libs = gnu_compiler_internal_libs
+    internal_libs = ()
 
     def __init__(self, exelist, version, is_cross, exe_wrap, is_64):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
