@@ -20,7 +20,7 @@ import pickle
 from mesonbuild import build
 from mesonbuild import environment
 from mesonbuild.dependencies import ExternalProgram
-from mesonbuild import mesonlib
+from mesonbuild.mesonlib import MesonException
 from mesonbuild import mlog
 
 import time, datetime, multiprocessing, json
@@ -105,7 +105,36 @@ def buildparser():
     return parser
 
 
-class TestException(mesonlib.MesonException):
+def returncode_to_status(retcode):
+    # Note: We can't use `os.WIFSIGNALED(result.returncode)` and the related
+    # functions here because the status returned by subprocess is munged. It
+    # returns a negative value if the process was killed by a signal rather than
+    # the raw status returned by `wait()`. Also, If a shell sits between Meson
+    # the the actual unit test that shell is likely to convert a termination due
+    # to a signal into an exit status of 128 plus the signal number.
+    if retcode < 0:
+        signum = -retcode
+        try:
+            signame = signal.Signals(signum).name
+        except ValueError:
+            signame = 'SIGinvalid'
+        return '(killed by signal %d %s)' % (signum, signame)
+
+    if retcode <= 128:
+        return '(exit status %d)' % (retcode,)
+
+    signum = retcode - 128
+    try:
+        signame = signal.Signals(signum).name
+    except ValueError:
+        signame = 'SIGinvalid'
+    return '(exit status %d or signal %d %s)' % (retcode, signum, signame)
+
+def env_tuple_to_str(env):
+    return ''.join(["%s='%s' " % (k, v) for k, v in env])
+
+
+class TestException(MesonException):
     pass
 
 
@@ -136,7 +165,8 @@ class TestRun:
         if self.cmd is None:
             res += 'NONE\n'
         else:
-            res += '%s%s\n' % (''.join(["%s='%s' " % (k, v) for k, v in self.env.items()]), ' ' .join(self.cmd))
+            test_only_env = set(self.env.items()) - set(os.environ.items())
+            res += '{}{}\n'.format(env_tuple_to_str(test_only_env), ' '.join(self.cmd))
         if self.stdo:
             res += '--- stdout ---\n'
             res += self.stdo
@@ -231,11 +261,27 @@ class SingleTestRunner:
                 self.test.timeout = None
             return self._run_cmd(wrap + cmd + self.test.cmd_args + self.options.test_args)
 
+    @staticmethod
+    def _substring_in_list(substr, strlist):
+        for s in strlist:
+            if substr in s:
+                return True
+        return False
+
     def _run_cmd(self, cmd):
         starttime = time.time()
 
         if len(self.test.extra_paths) > 0:
             self.env['PATH'] = os.pathsep.join(self.test.extra_paths + ['']) + self.env['PATH']
+            if self._substring_in_list('wine', cmd):
+                wine_paths = ['Z:' + p for p in self.test.extra_paths]
+                wine_path = ';'.join(wine_paths)
+                # Don't accidentally end with an `;` because that will add the
+                # current directory and might cause unexpected behaviour
+                if 'WINEPATH' in self.env:
+                    self.env['WINEPATH'] = wine_path + ';' + self.env['WINEPATH']
+                else:
+                    self.env['WINEPATH'] = wine_path
 
         # If MALLOC_PERTURB_ is not set, or if it is set to an empty value,
         # (i.e., the test or the environment don't explicitly set it), set
@@ -304,7 +350,8 @@ class SingleTestRunner:
                 subprocess.call(['taskkill', '/F', '/T', '/PID', str(p.pid)])
             else:
                 try:
-                    os.killpg(os.getpgid(p.pid), signal.SIGKILL)
+                    # Kill the process group that setsid() created.
+                    os.killpg(p.pid, signal.SIGKILL)
                 except ProcessLookupError:
                     # Sometimes (e.g. with Wine) this happens.
                     # There's nothing we can do (maybe the process
@@ -404,11 +451,16 @@ class TestHarness:
         num = '%s%d/%d' % (startpad, i + 1, len(tests))
         padding1 = ' ' * (38 - len(name))
         padding2 = ' ' * (8 - len(result.res.value))
-        result_str = '%s %s  %s%s%s%5.2f s' % \
-            (num, name, padding1, result.res.value, padding2, result.duration)
+        status = ''
+
+        if result.res is TestResult.FAIL:
+            status = returncode_to_status(result.returncode)
+        result_str = '%s %s  %s%s%s%5.2f s %s' % \
+            (num, name, padding1, result.res.value, padding2, result.duration,
+             status)
         if not self.options.quiet or result.res is not TestResult.OK:
             if result.res is not TestResult.OK and mlog.colorize_console:
-                if result.res is TestResult.FAIL or result.res is TestResult.TIMEOUT:
+                if result.res in (TestResult.FAIL, TestResult.TIMEOUT):
                     decorator = mlog.red
                 elif result.res is TestResult.SKIP:
                     decorator = mlog.yellow
@@ -560,6 +612,8 @@ TIMEOUT: %4d
 
         self.logfile.write('Log of Meson test suite run on %s\n\n'
                            % datetime.datetime.now().isoformat())
+        inherit_env = env_tuple_to_str(os.environ.items())
+        self.logfile.write('Inherited environment: {}\n\n'.format(inherit_env))
 
     @staticmethod
     def get_wrapper(options):
