@@ -20,9 +20,9 @@
 # This file is basically a reimplementation of
 # http://cgit.freedesktop.org/libreoffice/core/commit/?id=3213cd54b76bc80a6f0516aac75a48ff3b2ad67c
 
-import os, sys, subprocess
+import os, sys
 from .. import mesonlib
-from ..mesonlib import MesonException, Popen_safe
+from ..mesonlib import Popen_safe
 import argparse
 
 parser = argparse.ArgumentParser()
@@ -80,7 +80,7 @@ def osx_syms(libfilename, outfilename):
         if 'LC_ID_DYLIB' in val:
             match = i
             break
-    result = [arr[match+2], arr[match+5]] # Libreoffice stores all 5 lines but the others seem irrelevant.
+    result = [arr[match + 2], arr[match + 5]] # Libreoffice stores all 5 lines but the others seem irrelevant.
     pnm, output = Popen_safe(['nm', '-g', '-P', libfilename])[0:2]
     if pnm.returncode != 0:
         raise RuntimeError('nm does not work.')
