@@ -16,34 +16,43 @@ import json
 import sys, os
 import configparser
 import shutil
+import argparse
 
 from glob import glob
 
 from .wrap import API_ROOT, open_wrapdburl
 
-help_templ = '''This program allows you to manage your Wrap dependencies
-using the online wrap database http://wrapdb.mesonbuild.com.
+from .. import mesonlib
 
-Run this command in your top level source directory.
+def add_arguments(parser):
+    subparsers = parser.add_subparsers(title='Commands', dest='command')
+    subparsers.required = True
 
-Usage:
+    p = subparsers.add_parser('list', help='show all available projects')
+    p.set_defaults(wrap_func=list_projects)
 
-%s <command> [options]
+    p = subparsers.add_parser('search', help='search the db by name')
+    p.add_argument('name')
+    p.set_defaults(wrap_func=search)
 
-Commands:
+    p = subparsers.add_parser('install', help='install the specified project')
+    p.add_argument('name')
+    p.set_defaults(wrap_func=install)
 
- list - show all available projects
- search - search the db by name
- install - install the specified project
- update - update the project to its newest available release
- info - show available versions of a project
- status - show installed and available versions of your projects
+    p = subparsers.add_parser('update', help='update the project to its newest available release')
+    p.add_argument('name')
+    p.set_defaults(wrap_func=update)
 
-'''
+    p = subparsers.add_parser('info', help='show available versions of a project')
+    p.add_argument('name')
+    p.set_defaults(wrap_func=info)
 
+    p = subparsers.add_parser('status', help='show installed and available versions of your projects')
+    p.set_defaults(wrap_func=status)
 
-def print_help():
-    print(help_templ % sys.argv[0])
+    p = subparsers.add_parser('promote', help='bring a subsubproject up to the master project')
+    p.add_argument('project_path')
+    p.set_defaults(wrap_func=promote)
 
 def get_result(urlstring):
     u = open_wrapdburl(urlstring)
@@ -60,12 +69,13 @@ def get_projectlist():
     projects = jd['projects']
     return projects
 
-def list_projects():
+def list_projects(options):
     projects = get_projectlist()
     for p in projects:
         print(p)
 
-def search(name):
+def search(options):
+    name = options.name
     jd = get_result(API_ROOT + 'query/byname/' + name)
     for p in jd['projects']:
         print(p)
@@ -76,7 +86,8 @@ def get_latest_version(name):
     revision = jd['revision']
     return branch, revision
 
-def install(name):
+def install(options):
+    name = options.name
     if not os.path.isdir('subprojects'):
         print('Subprojects dir not found. Run this script in your source root directory.')
         sys.exit(1)
@@ -104,7 +115,8 @@ def get_current_version(wrapfile):
     revision = int(arr[-2])
     return branch, revision, cp['directory'], cp['source_filename'], cp['patch_filename']
 
-def update(name):
+def update(options):
+    name = options.name
     if not os.path.isdir('subprojects'):
         print('Subprojects dir not found. Run this command in your source root directory.')
         sys.exit(1)
@@ -132,7 +144,8 @@ def update(name):
         f.write(data)
     print('Updated', name, 'to branch', new_branch, 'revision', new_revision)
 
-def info(name):
+def info(options):
+    name = options.name
     jd = get_result(API_ROOT + 'projects/' + name)
     versions = jd['versions']
     if not versions:
@@ -142,10 +155,41 @@ def info(name):
     for v in versions:
         print(' ', v['branch'], v['revision'])
 
-def status():
+def do_promotion(from_path, spdir_name):
+    if os.path.isfile(from_path):
+        assert(from_path.endswith('.wrap'))
+        shutil.copy(from_path, spdir_name)
+    elif os.path.isdir(from_path):
+        sproj_name = os.path.basename(from_path)
+        outputdir = os.path.join(spdir_name, sproj_name)
+        if os.path.exists(outputdir):
+            sys.exit('Output dir %s already exists. Will not overwrite.' % outputdir)
+        shutil.copytree(from_path, outputdir, ignore=shutil.ignore_patterns('subprojects'))
+
+def promote(options):
+    argument = options.project_path
+    path_segment, subproject_name = os.path.split(argument)
+    spdir_name = 'subprojects'
+    sprojs = mesonlib.detect_subprojects(spdir_name)
+    if subproject_name not in sprojs:
+        sys.exit('Subproject %s not found in directory tree.' % subproject_name)
+    matches = sprojs[subproject_name]
+    if len(matches) == 1:
+        do_promotion(matches[0], spdir_name)
+        return
+    if path_segment == '':
+        print('There are many versions of %s in tree. Please specify which one to promote:\n' % subproject_name)
+        for s in matches:
+            print(s)
+        sys.exit(1)
+    system_native_path_argument = argument.replace('/', os.sep)
+    if system_native_path_argument in matches:
+        do_promotion(argument, spdir_name)
+
+def status(options):
     print('Subproject status')
     for w in glob('subprojects/*.wrap'):
-        name = os.path.split(w)[1][:-5]
+        name = os.path.basename(w)[:-5]
         try:
             (latest_branch, latest_revision) = get_latest_version(name)
         except Exception:
@@ -162,39 +206,10 @@ def status():
             print('', name, 'not up to date. Have %s %d, but %s %d is available.' % (current_branch, current_revision, latest_branch, latest_revision))
 
 def run(args):
-    if not args or args[0] == '-h' or args[0] == '--help':
-        print_help()
-        return 0
-    command = args[0]
-    args = args[1:]
-    if command == 'list':
-        list_projects()
-    elif command == 'search':
-        if len(args) != 1:
-            print('Search requires exactly one argument.')
-            return 1
-        search(args[0])
-    elif command == 'install':
-        if len(args) != 1:
-            print('Install requires exactly one argument.')
-            return 1
-        install(args[0])
-    elif command == 'update':
-        if len(args) != 1:
-            print('update requires exactly one argument.')
-            return 1
-        update(args[0])
-    elif command == 'info':
-        if len(args) != 1:
-            print('info requires exactly one argument.')
-            return 1
-        info(args[0])
-    elif command == 'status':
-        status()
-    else:
-        print('Unknown command', command)
-        return 1
-    return 0
+    parser = argparse.ArgumentParser(prog='wraptool')
+    add_arguments(parser)
+    options = parser.parse_args(args)
+    options.wrap_func(options)
 
 if __name__ == '__main__':
     sys.exit(run(sys.argv[1:]))
